/* Copyright 2000 John Fremlin
   Distributed under GPL, see file COPYING
*/

#define _GNU_SOURCE

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <signal.h>
#include <err.h>
#include <time.h>

#include <sys/utsname.h>
#include "config.h"

#ifndef USE_UTMP
#define USE_UTMP 1
#endif

#include <errno.h>
#include <unistd.h>

#if USE_UTMP
#include <utmp.h>
#endif

#ifdef HAVE_CURSES_H
#include <curses.h>
#endif
#ifdef HAVE_NCURSES_H
#include <ncurses.h>
#endif
#ifdef HAVE_TERM_H
#include <term.h>
#endif

#include "system.h"

#ifndef LOGIN_MESSAGE
#define LOGIN_MESSAGE _("%n login: ")
#endif

#ifdef HAVE_CURSES
inline static
void
xtputs(const char*str)
{
  if(!str)
    return;
  
  if(tputs(str,STDOUT_FILENO,putchar)==ERR)
    warn(_("can't tputs"));
}
#endif


inline static
void
clear_tty()
{
#ifdef HAVE_CURSES
  int error;

  if(setupterm(0,STDOUT_FILENO,&error)==ERR){
    warn(_("can't setupterm"));
    return;
  }
  xtputs(reset_1string);
  xtputs(reset_2string);
  xtputs(clear_screen);
  
#endif  
}


static char*
shorten_ttyname(char*name)
{
  if(!name)
    return 0;
  
  if(!strncmp(name,_PATH_DEV,strlen(_PATH_DEV)))
    name+=strlen(_PATH_DEV);
  
  return name;
}

inline
static
void
utmp_admin(char* ttyname,const char*hostname)
{
#if USE_UTMP
  struct utmp* utp;
  struct utmp entry;
  char*tty=0;
  pid_t pid;
  
  if(utmpname(_PATH_UTMP)==-1)
    {
      warn("utmpname");
      return;
    }

  tty = shorten_ttyname(ttyname);
  
  setutent();
  
  pid=getpid();
  while((utp=getutent()))
    if(utp->ut_pid==pid)
      {
	memcpy(&entry,utp,sizeof(entry));
	goto got_entry;
      }

  memset(&entry,0,sizeof(entry));
  
 got_entry:  

  strncpy(entry.ut_user, "LOGIN", sizeof(entry.ut_user));
  if(tty)
    strncpy(entry.ut_line, tty, sizeof(entry.ut_line));
  strncpy(entry.ut_host, hostname, sizeof(entry.ut_host));
  
  entry.ut_time = time(0);
  entry.ut_type = LOGIN_PROCESS;
  entry.ut_pid = pid;
  entry.ut_session = pid;
  
  if(!pututline(&entry))
      warn("pututline");

  endutent();
      
  updwtmp(_PATH_WTMP,&entry);

#endif
}

static int
display_formatted(FILE* stream)
{
  int c;
  static struct utsname uts;
  static int uts_initted;
  static char*tty = 0;
  
  if(!uts_initted && uname(&uts)==-1)
    warn("uname");
  uts_initted = 1;

  while((c=fgetc(stream))!=EOF) {
    if(c=='%'){
      c=fgetc(stream);
      switch(c){
      case 'l':
	if(!tty) {
	  tty = ttyname(0);
	  if(!tty)
	    tty = (char*)_("not a tty");
	  tty = shorten_ttyname(tty);
	}
	printf("%s",tty);
	break;
      case 'd':
	{
	  time_t t;
	  char*str;
	  if(time(&t)==-1 || !(str = ctime(&t)))
	    warn(_("can't get time"));
	  else {
	    char*nl = strchr(str,'\n');
	    if(nl)*nl = 0;
	    printf("%s",str);
	  }
	}
	
	break;
#define show_uts(x,p) \
case x: \
	printf("%.*s",(int)sizeof(uts.p),uts.p); \
        break

	show_uts('s',sysname);
	show_uts('r',release);
	show_uts('v',version);
	show_uts('m',machine);
	show_uts('n',nodename);
	show_uts('o',domainname);

#undef show_uts
      case EOF:
	warnx(_("EOF in escape sequence"));
	break;

      case '%':
	putchar('%');
	break;
      default:
	putchar('%');
	putchar(c);
	break;
      }
    }
    else
      putchar(c);
  }

  fflush(stdout);
  fclose(stream);
  
  return 0;
}

static int
display(const char*file)
{
  FILE* stream = fopen(file,"r");

  if(!stream) {
    warn(_("open %s"),file);
    return 0;
  }

  return display_formatted(stream);
}


int print_prompt(void)
{
  FILE* stream = fmemopen((char*)LOGIN_MESSAGE,strlen(LOGIN_MESSAGE),"r");
  if(!stream)
    err(EXIT_FAILURE,"fmemopen");
  
  return display_formatted(stream);
}

char* read_username(void)
{
  char username[UT_NAMESIZE+1];
  char*ptr;
  
  if(!fgets(username,sizeof(username),stdin))
    {
      putc('\n',stderr);
      if(feof(stdin))
	warnx(_("EOF reading username"));
      else
	warn(_("unable to read username"));
      return 0;
    }
  for(ptr=username;*ptr;ptr++)
    {
      if(*ptr=='\n')
	{
	  *ptr=0;
	  goto polished_off_name;
	}
    }
  warnx(_("username too long"));
  return 0;
 polished_off_name:
  return strdup(username);
}

int
main(int argc,char*argv[])
{
  char*username;
  gettext_init();

  utmp_admin(ttyname(0),"");

  clear_tty();
  
  if(display(_PATH_ETC_ISSUE))
    return 1;

  if(print_prompt())
    return 2;
  
  username = read_username();
  if(!username)
    return 3;
  
  if(!argv[1]){
    execl(LOGIN_PROGRAM,LOGIN_PROGRAM,username,0);
    warn("unable to exec default login program %s",LOGIN_PROGRAM);
  }
  else {
    int i;
    for(i=0;argv[i+1];i++)
      argv[i] = argv[i+1];
    argv[i] = username;
    execv(argv[0],argv);
    warn("unable to exec %s",argv[0]);
  }
  return 4;
}
